# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Logit processor for supporting watermarking in HF model."""

from collections.abc import Sequence

import torch
import transformers

from . import hashing_function


def update_scores(
    scores: torch.FloatTensor,
    g_values: torch.FloatTensor,
) -> torch.FloatTensor:
  """Updates scores using the g values.

  We assume that the scores are in the log space.
  Args:
    scores: Scores (batch_size, vocab_size).
    g_values: G valus (batch_size, vocab_size, depth).

  Returns:
    Updated scores (batch_size, vocab_size).
  """
  _, _, depth = g_values.shape
  device = scores.device

  probs = torch.softmax(scores, dim=1)

  for i in range(depth):
    g_values_at_depth = g_values[:, :, i]
    g_mass_at_depth = (g_values_at_depth * probs).sum(axis=1, keepdims=True)
    probs = probs * (1 + g_values_at_depth - g_mass_at_depth)

  log_probs = torch.log(probs)
  log_probs = torch.where(
      torch.isfinite(log_probs), log_probs, torch.tensor(-1e12, device=device)
  )
  return log_probs


def update_scores_distortionary(
    scores: torch.FloatTensor,
    g_values: torch.FloatTensor,
    num_leaves: int,
) -> torch.FloatTensor:
  """Update scores using the g values for distortionary tournament watermarking.

  We assume that the scores are in the log space.
  Args:
    scores: Scores (batch_size, vocab_size).
    g_values: G valus (batch_size, vocab_size, depth).
    num_leaves: Number of leaves per node in the tournament tree.

  Returns:
    Updated scores (batch_size, vocab_size).
  """
  _, _, depth = g_values.shape
  device = scores.device

  probs = torch.softmax(scores, dim=1)

  for i in range(depth):
    g_values_at_depth = g_values[:, :, i]
    g_mass_at_depth = (g_values_at_depth * probs).sum(axis=1, keepdims=True)
    coeff_not_in_g = (1 - g_mass_at_depth)**(num_leaves - 1)
    coeff_in_g = (1 - (1 - g_mass_at_depth)**(num_leaves)) / g_mass_at_depth
    coeffs = torch.where(
        torch.logical_and(g_values_at_depth == 1, probs > 0),
        coeff_in_g, coeff_not_in_g)
    probs = probs * coeffs

  log_probs = torch.log(probs)
  log_probs = torch.where(
      torch.isfinite(log_probs), log_probs, torch.tensor(-1e12, device=device)
  )
  return log_probs


class SynthIDState:
  """SynthID watermarking state."""

  def __init__(
      self,
      batch_size: int,
      ngram_len: int,
      context_history_size: int,
      device: torch.device,
  ):
    """Initializes the state.

    Args:
      batch_size: Batch size.
      ngram_len: Ngram length.
      context_history_size: Size of the tensor to keep track of seen contexts.
      device: Device to use.
    """
    self.context = torch.zeros(
        (batch_size, ngram_len - 1),
        dtype=torch.int64,
        device=device,
    )
    self.context_history = torch.zeros(
        (batch_size, context_history_size),
        dtype=torch.int64,
        device=device,
    )
    self.num_calls = 0


class SynthIDLogitsProcessor(transformers.LogitsProcessor):
  """SynthID watermarking logits processor.

  Logits processor updates the provided scores based on the binary g values
  assigned to each possible ngram and watermarking key combination hashed into
  an int64 keys.

  A random sampling table is pre-computed and modulo table size is applied to
  map from ngram keys (int64) to g values.
  """

  def __init__(
      self,
      *,
      ngram_len: int,
      keys: Sequence[int],
      sampling_table_size: int,
      sampling_table_seed: int,
      context_history_size: int,
      temperature: float,
      top_k: int,
      device: torch.device,
      skip_first_ngram_calls: bool = False,
      apply_top_k: bool = True,
      num_leaves: int = 2
  ):
    """Initializes the logits processor.

    Args:
      ngram_len: Ngram length.
      keys: A sequence of watermarking keys, one for each depth.
      sampling_table_size: Size of the sampling table.
      sampling_table_seed: Random seed to generate the sampling table.
      context_history_size: Size of the tensor to keep track of seen contexts.
      temperature: Temperature to use for scaling the scores.
      top_k: Top k to use for sampling the scores.
      device: Device to use.
      skip_first_ngram_calls: Whether to skip first ngram calls.
      apply_top_k: Whether to apply top k to the scores.
      num_leaves: Number of leaves per node in the tournament tree.
    """
    self.ngram_len = ngram_len
    self.keys = torch.tensor(keys, device=device)

    generator = torch.Generator(device=device).manual_seed(sampling_table_seed)
    # A random sampling table is pre-computed and modulo table size is applied
    # to map from a hash of ngram keys to g values, this is similar to the
    # hashtable implementation used in
    # https://github.com/facebookresearch/three_bricks. We note that the
    # hashing employed in this repository is different from that used to
    # watermark the Gemini App, and hence the detectors trained based on the
    # hashing in this repository will not transfer to text generated by
    # the Gemini App.
    self.sampling_table = torch.randint(
        low=0,
        high=2,
        size=(sampling_table_size,),
        generator=generator,
        device=device,
    )
    self.context_history_size = context_history_size
    self.device = device
    self.state = None
    self.skip_first_ngram_calls = skip_first_ngram_calls
    self.apply_top_k = apply_top_k

    # Check validity of temperature.
    if not (isinstance(temperature, float) and temperature > 0):
      except_msg = (
          f"`temperature` (={temperature}) has to be a strictly positive float,"
          " otherwise your next token scores will be invalid."
      )
      if isinstance(temperature, float) and temperature == 0.0:
        except_msg += (
            " If you're looking for greedy decoding strategies, set"
            " `do_sample=False`."
        )
      raise ValueError(except_msg)

    self.temperature = temperature

    self._num_leaves = num_leaves

    # Check validity of top_k.
    if not (isinstance(top_k, int) and top_k > 1):
      raise ValueError(f"`top_k` has to be > 1, but is {top_k}")

    self.top_k = top_k

  def _init_state(self, batch_size: int):
    """Initializes the state."""
    self.state = SynthIDState(
        batch_size=batch_size,
        ngram_len=self.ngram_len,
        context_history_size=self.context_history_size,
        device=self.device,
    )

  @torch.no_grad
  def __call__(
      self,
      input_ids: torch.LongTensor,
      scores: torch.FloatTensor,
  ) -> tuple[torch.FloatTensor, torch.LongTensor]:
    raise NotImplementedError(
        "__call__ is not implemented for watermarking logits processor."
    )

  @torch.no_grad
  def watermarked_call(
      self,
      input_ids: torch.LongTensor,
      scores: torch.FloatTensor,
  ) -> tuple[torch.FloatTensor, torch.LongTensor, torch.FloatTensor]:
    """Calls the logits processor statefully.

    This function computes top_k internally and returns the indices mapping
    from top_k scores to dense scores.

    Args:
      input_ids: Input token ids (batch_size, inputs_len).
      scores: Scores (batch_size, vocab_size).

    Returns:
      Tuple of
        Watermarked updated scores (batch_size, top_k)
        Top k indices (batch_size, top_k).
        original scores for perplexity calculations (batch_size, top_k)
    """
    self._check_input_ids_shape(input_ids)
    scores_processed = scores / self.temperature
    top_k_result = torch.topk(scores_processed, k=self.top_k, dim=1)
    batch_size, vocab_size = scores.shape

    if self.apply_top_k:
      scores_top_k = top_k_result.values
      # scores_top_k shape [batch_size, top_k]
      top_k_indices = top_k_result.indices
      # top_k_indices shape [batch_size, top_k]
    else:
      scores_top_k = scores_processed
      top_k_indices = torch.stack([
          torch.arange(vocab_size, device=self.device)
          for _ in range(batch_size)
      ])

    device = scores.device
    if device != self.device:
      raise ValueError(
          "SynthIDLogitsProcessor received inputs with unexpected device.",
      )

    if self.state is None:
      # Initialize watermarking state if it does not exist.
      self._init_state(batch_size)
    else:
      # Append last input id (which is the input id added in last call) to the
      # previous context so we have the context to be used for current
      # watermarking.
      self.state.context = torch.concat(
          (self.state.context, input_ids[:, -1:]),
          dim=1,
      )
      self.state.context = self.state.context[:, 1:]

    assert self.state is not None

    self.state.num_calls += 1

    # Don't watermark the first ngram_len - 1 tokens if set.
    if self.skip_first_ngram_calls and self.state.num_calls < self.ngram_len:
      return scores_top_k, top_k_indices, scores_top_k

    # 2. Generate random keys for each ngram key combination.
    ngram_keys, hash_result_with_just_context = self._compute_keys(
        self.state.context, top_k_indices
    )
    # ngram_keys shape [batch_size, top_k, depth]

    # 3. Sample g values.
    g_values = self.sample_g_values(ngram_keys)
    # g_values shape [batch_size, top_k, depth]

    # 4. Modify scores.
    if self._num_leaves == 2:
      updated_scores = update_scores(scores_top_k, g_values)
    else:
      updated_scores = update_scores_distortionary(
          scores_top_k, g_values, self._num_leaves
      )
    # updated scores shape [batch_size, top_k]

    # 5. Check if the current watermarking context was previously used, if
    # yes skip watermarking.
    hash_result_with_just_context = hash_result_with_just_context[:, None]
    is_repeated_context = (
        self.state.context_history == hash_result_with_just_context
    ).any(
        dim=1,
        keepdim=True,
    )
    self.state.context_history = torch.concat(
        (hash_result_with_just_context, self.state.context_history),
        dim=1,
    )[:, :-1]

    updated_watermarked_scores = torch.where(
        is_repeated_context,
        input=scores_top_k,
        other=updated_scores,
    )
    return updated_watermarked_scores, top_k_indices, scores_top_k

  def compute_ngram_keys(
      self,
      ngrams: torch.LongTensor,
  ) -> torch.LongTensor:
    """Computes random keys for each ngram and depth.

    Args:
      ngrams: Ngrams (batch_size, num_ngrams, ngram_len).

    Returns:
      ngram keys (batch_size, num_ngrams, depth).
    """
    if len(ngrams.shape) != 3:
      raise ValueError(
          "Ngrams should be of shape (batch_size, num_ngrams, ngram_len), but"
          f" is {ngrams.shape}"
      )
    if ngrams.shape[2] != self.ngram_len:
      raise ValueError(
          "Ngrams should be of shape (batch_size, num_ngrams, ngram_len),"
          f" where ngram_len is {self.ngram_len}, but is {ngrams.shape}"
      )
    batch_size, _, _ = ngrams.shape

    hash_result = torch.ones(batch_size, device=self.device, dtype=torch.long)
    # hash_result shape [batch_size,]
    # ngrams shape [batch_size, num_ngrams, ngram_len]
    hash_result = torch.vmap(
        hashing_function.accumulate_hash, in_dims=(None, 1), out_dims=1
    )(hash_result, ngrams)
    # hash_result shape [batch_size, num_ngrams]

    keys = self.keys[None, None, :, None]
    # hash_result shape [batch_size, num_ngrams]
    # keys shape [1, 1, depth, 1]
    hash_result = torch.vmap(
        hashing_function.accumulate_hash, in_dims=(None, 2), out_dims=2
    )(hash_result, keys)
    # hash_result shape [batch_size, num_ngrams, depth]

    return hash_result

  def _compute_keys(
      self,
      n_minus_1_grams: torch.LongTensor,
      indices: torch.LongTensor,
  ) -> tuple[torch.LongTensor, torch.LongTensor]:
    """Computes random keys for each ngram and depth.

    Args:
      n_minus_1_grams: Ngrams (batch_size, ngram_len - 1).
      indices: indices of the continuations (batch_size, num_indices)

    Returns:
      Ngram keys (batch_size, num_indices, depth).
    """
    batch_size, _ = n_minus_1_grams.shape

    hash_result = torch.ones(batch_size, device=self.device, dtype=torch.long)
    # First hash n_minus_1 gram, for each batch entry we have a single
    # n_minus_1 gram context.
    # hash_result shape [batch_size]
    # n_minus_1_gram shape [batch_size, ngram_len - 1]
    hash_result_with_just_context = hashing_function.accumulate_hash(
        hash_result, n_minus_1_grams
    )
    # hash_result shape [batch_size,]
    # Indices is of shape [batch_size, num_indices], so we make it
    # [batch_size, num_indices, 1] so we can vmap over num_indices dim.
    hash_result = torch.vmap(
        hashing_function.accumulate_hash, in_dims=(None, 1), out_dims=1
    )(hash_result_with_just_context, indices[:, :, None])
    # hash_result shape [batch_size, num_indices]
    # Basically we have a hash for each batch entry and each indices
    # Now we add watermarking keys to this hash.
    # keys are of shape [depth,]
    # We add batch, num_indices and data dimension to this making it
    # [1, 1, depth, 1].
    # So we can vmap over the depth dimension for compute_hash
    keys = self.keys[None, None, :, None]
    hash_result = torch.vmap(
        hashing_function.accumulate_hash, in_dims=(None, 2), out_dims=2
    )(hash_result, keys)
    # hash_result shape should be [batch_size, num_indices, depth]
    return hash_result, hash_result_with_just_context

  def sample_g_values(self, ngram_keys: torch.LongTensor) -> torch.LongTensor:
    """Samples g values from Bernoulli distribution.

    It is not possible to pass random keys in a vectorized way in torch. Instead
    we pre-compute a random sampling table, and use apply modulo table size to
    map from ngram keys (int64) to g values.

    Args:
      ngram_keys: Random keys (batch_size, num_ngrams, depth).

    Returns:
      G values (batch_size, num_ngrams, depth).
    """
    (sampling_table_size,) = self.sampling_table.shape
    sampling_table = self.sampling_table.reshape((1, 1, sampling_table_size))
    ngram_keys = ngram_keys % sampling_table_size
    return torch.take_along_dim(sampling_table, indices=ngram_keys, dim=2)

  def _check_input_ids_shape(self, input_ids: torch.LongTensor):
    """Checks the shape of input ids."""
    if len(input_ids.shape) != 2:
      raise ValueError(
          "Input ids should be of shape (batch_size, input_len), but is"
          f" {input_ids.shape}"
      )

  def compute_g_values(
      self,
      input_ids: torch.LongTensor,
  ) -> torch.LongTensor:
    """Computes g values for each ngram from the given sequence of tokens.

    Args:
      input_ids: Input token ids (batch_size, input_len).

    Returns:
      G values (batch_size, input_len - (ngram_len - 1), depth).
    """
    self._check_input_ids_shape(input_ids)
    ngrams = input_ids.unfold(dimension=1, size=self.ngram_len, step=1)
    ngram_keys = self.compute_ngram_keys(ngrams)
    return self.sample_g_values(ngram_keys)

  def compute_context_repetition_mask(
      self,
      input_ids: torch.LongTensor,
  ) -> torch.LongTensor:
    """Computes repetition mask.

    0 and 1 stand for repeated and not repeated context n-1 grams respectively.

    Args:
      input_ids: Input token ids (batch_size, input_len).

    Returns:
      Repetitions mask (batch_size, input_len - (ngram_len - 1)).
    """
    self._check_input_ids_shape(input_ids)
    batch_size, _ = input_ids.shape
    state = SynthIDState(
        batch_size=batch_size,
        ngram_len=self.ngram_len,
        context_history_size=self.context_history_size,
        device=self.device,
    )
    contexts = input_ids[:, :-1].unfold(
        dimension=1,
        size=self.ngram_len - 1,
        step=1,
    )
    _, num_contexts, _ = contexts.shape

    are_repeated_contexts = []
    for i in range(num_contexts):
      context = contexts[:, i, :]
      hash_result = torch.ones(batch_size, device=self.device, dtype=torch.long)
      context_hash = hashing_function.accumulate_hash(hash_result, context)[
          :, None
      ]
      is_repeated_context = (state.context_history == context_hash).any(
          dim=1,
          keepdim=True,
      )
      are_repeated_contexts.append(is_repeated_context)
      state.context_history = torch.concat(
          (context_hash, state.context_history),
          dim=1,
      )[:, :-1]
    are_repeated_contexts = torch.concat(are_repeated_contexts, dim=1)

    return torch.logical_not(are_repeated_contexts)

  def compute_eos_token_mask(
      self,
      input_ids: torch.LongTensor,
      eos_token_id: int,
  ) -> torch.LongTensor:
    """Computes repetitions mask.

    1 stands for ngrams that don't contain EOS tokens and vice versa.

    Args:
      input_ids: Input token ids (batch_size, input_len).
      eos_token_id: EOS token ID.

    Returns:
      EOS token mask (batch_size, input_len).
    """
    self._check_input_ids_shape(input_ids)
    noneos_masks = []
    all_eos_equated = input_ids == eos_token_id
    for eos_equated in all_eos_equated:
      nonzero_idx = torch.nonzero(eos_equated)
      noneos_mask = torch.ones_like(eos_equated)
      if nonzero_idx.shape[0] != 0:
        noneos_mask[nonzero_idx[0][0] :] = 0
      noneos_masks.append(noneos_mask)
    return torch.stack(noneos_masks, dim=0)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Bayesian detector class."""

import abc
from collections.abc import Mapping, Sequence
import enum
import functools
import gc
from typing import Any
import flax.linen as nn
import jax
import jax.numpy as jnp
from jaxtyping import PyTree  # pylint: disable=g-importing-member
import numpy as np
import optax
from sklearn import model_selection
import torch
import tqdm

from . import logits_processing


def pad_to_len(
    arr: torch.tensor,
    target_len: int,
    *,
    left_pad: bool,
    eos_token: int,
    device: torch.device,
) -> torch.tensor:
  """Pad or truncate array to given length."""
  if arr.shape[1] < target_len:
    shape_for_ones = list(arr.shape)
    shape_for_ones[1] = target_len - shape_for_ones[1]
    padded = (
        torch.ones(
            shape_for_ones,
            device=device,
            dtype=torch.long,
        )
        * eos_token
    )
    if not left_pad:
      return torch.concatenate((arr, padded), dim=1)
    else:
      return torch.concatenate((padded, arr), dim=1)
  else:
    return arr[:, :target_len]


def filter_and_truncate(
    outputs: torch.tensor,
    truncation_length: int | None,
    eos_token_mask: torch.tensor,
) -> torch.tensor:
  """Filter and truncate outputs to given length.

  Args:
   outputs: output tensor of shape [batch_size, output_len]
   truncation_length: Length to truncate the final output. If None, then no
     truncation is applied.
   eos_token_mask: EOS token mask of shape [batch_size, output_len]

  Returns:
   output tensor of shape [batch_size, truncation_length].
  """
  if truncation_length:
    outputs = outputs[:, :truncation_length]
    truncation_mask = torch.sum(eos_token_mask, dim=1) >= truncation_length
    return outputs[truncation_mask, :]
  return outputs


def process_outputs_for_training(
    all_outputs: Sequence[torch.Tensor],
    logits_processor: logits_processing.SynthIDLogitsProcessor,
    tokenizer: Any,
    *,
    pos_truncation_length: int | None,
    neg_truncation_length: int | None,
    max_length: int,
    is_cv: bool,
    is_pos: bool,
    torch_device: torch.device,
) -> tuple[Sequence[torch.tensor], Sequence[torch.tensor]]:
  """Process raw model outputs into format understandable by the detector.

  Args:
   all_outputs: sequence of outputs of shape [batch_size, output_len].
   logits_processor: logits processor used for watermarking.
   tokenizer: tokenizer used for the model.
   pos_truncation_length: Length to truncate the watermarked outputs. If None,
     then no truncation is applied.
   neg_truncation_length: Length to truncate the unwatermarked outputs. If None,
     then no truncation is applied.
   max_length: Length to pad truncated outputs so that all processed entries.
     have same shape.
   is_cv: Process given outputs for cross validation.
   is_pos: Process given outputs for positives.
   torch_device: torch device to use.

  Returns:
    Tuple of
      all_masks: list of masks of shape [batch_size, max_length].
      all_g_values: list of g_values of shape [batch_size, max_length, depth].
  """
  all_masks = []
  all_g_values = []
  for outputs in tqdm.tqdm(all_outputs):
    # outputs is of shape [batch_size, output_len].
    # output_len can differ from batch to batch.
    eos_token_mask = logits_processor.compute_eos_token_mask(
        input_ids=outputs,
        eos_token_id=tokenizer.eos_token_id,
    )
    if is_pos or is_cv:
      # filter with length for positives for both train and CV.
      # We also filter for length when CV negatives are processed.
      outputs = filter_and_truncate(
          outputs, pos_truncation_length, eos_token_mask
      )
    elif not is_pos and not is_cv:
      outputs = filter_and_truncate(
          outputs, neg_truncation_length, eos_token_mask
      )

    # If no filtered outputs skip this batch.
    if outputs.shape[0] == 0:
      continue

    # All outputs are padded to max-length with eos-tokens.
    outputs = pad_to_len(
        outputs,
        max_length,
        left_pad=False,
        eos_token=tokenizer.eos_token_id,
        device=torch_device,
    )
    # outputs shape [num_filtered_entries, max_length]

    eos_token_mask = logits_processor.compute_eos_token_mask(
        input_ids=outputs,
        eos_token_id=tokenizer.eos_token_id,
    )

    context_repetition_mask = logits_processor.compute_context_repetition_mask(
        input_ids=outputs,
    )

    # context_repetition_mask of shape [num_filtered_entries, max_length -
    # (ngram_len - 1)].
    context_repetition_mask = pad_to_len(
        context_repetition_mask,
        max_length,
        left_pad=True,
        eos_token=0,
        device=torch_device,
    )
    # We pad on left to get same max_length shape.
    # context_repetition_mask of shape [num_filtered_entries, max_length].
    combined_mask = context_repetition_mask * eos_token_mask

    g_values = logits_processor.compute_g_values(
        input_ids=outputs,
    )

    # g_values of shape [num_filtered_entries, max_length - (ngram_len - 1),
    # depth].
    g_values = pad_to_len(
        g_values, max_length, left_pad=True, eos_token=0, device=torch_device
    )

    # We pad on left to get same max_length shape.
    # g_values of shape [num_filtered_entries, max_length, depth].
    all_masks.append(combined_mask)
    all_g_values.append(g_values)
  return all_masks, all_g_values


@enum.unique
class ScoreType(enum.Enum):
  """Type of score returned by a WatermarkDetector.

  In all cases, larger score corresponds to watermarked text.
  """

  # Negative p-value where the p-value is the probability of observing equal or
  # stronger watermarking in unwatermarked text.
  NEGATIVE_P_VALUE = enum.auto()

  # Prob(watermarked | g-values).
  POSTERIOR = enum.auto()


class LikelihoodModel(abc.ABC):
  """Watermark likelihood model base class defining __call__ interface."""

  @abc.abstractmethod
  def __call__(self, g_values: jnp.ndarray) -> jnp.ndarray:
    """Computes likelihoods given g-values and a mask.

    Args:
      g_values: g-values (all are 0 or 1) of shape [batch_size, seq_len,
        watermarking_depth, ...].

    Returns:
      an array of shape [batch_size, seq_len, watermarking_depth] or
      [batch_size, seq_len, 1] corresponding to the likelihoods
      of the g-values given either the watermarked hypothesis or
      the unwatermarked hypothesis; i.e. either P(g|watermarked)
      or P(g|unwatermarked).
    """


class LikelihoodModelWatermarked(nn.Module, LikelihoodModel):
  """Watermarked likelihood model for binary-valued g-values.

  This takes in g-values and returns P(g_values|watermarked).
  """

  watermarking_depth: int
  params: Mapping[str, Mapping[str, Any]] | None = None

  def setup(self):
    """Initializes the model parameters."""

    def noise(seed, shape):
      return jax.random.normal(key=jax.random.PRNGKey(seed), shape=shape)

    self.beta = self.param(
        "beta",
        lambda *x: (
            -2.5 + 0.001 * noise(seed=0, shape=(1, 1, self.watermarking_depth))
        ),
    )
    self.delta = self.param(
        "delta",
        lambda *x: (
            0.001
            * noise(
                seed=0,
                shape=(1, 1, self.watermarking_depth, self.watermarking_depth),
            )
        ),
    )

  def l2_loss(self) -> jnp.ndarray:
    return jnp.einsum("ijkl->", self.delta**2)

  def _compute_latents(
      self, g_values: jnp.ndarray
  ) -> tuple[jnp.ndarray, jnp.ndarray]:
    """Computes the unique token probability distribution given g-values.

    Args:
      g_values: Pseudorandom function values of shape [batch_size, seq_len,
        watermarking_depth].

    Returns:
      p_one_unique_token and p_two_unique_tokens, both of shape
        [batch_size, seq_len, watermarking_depth]. p_one_unique_token[i,t,l]
        gives the probability of there being one unique token in a tournament
        match on layer l, on timestep t, for batch item i.
        p_one_unique_token[i,t,l] + p_two_unique_token[i,t,l] = 1.
    """
    # Tile g-values to produce feature vectors for predicting the latents
    # for each layer in the tournament; our model for the latents psi is a
    # logistic regression model psi = sigmoid(delta * x + beta).
    x = jnp.repeat(
        jnp.expand_dims(g_values, axis=-2), self.watermarking_depth, axis=-2
    )  # [batch_size, seq_len, watermarking_depth, watermarking_depth]

    x = jnp.tril(
        x, k=-1
    )  # mask all elements above -1 diagonal for autoregressive factorization

    logits = (
        jnp.einsum("ijkl,ijkl->ijk", self.delta, x) + self.beta
    )  # [batch_size, seq_len, watermarking_depth]

    p_two_unique_tokens = jax.nn.sigmoid(logits)
    p_one_unique_token = 1 - p_two_unique_tokens
    return p_one_unique_token, p_two_unique_tokens

  def __call__(self, g_values: jnp.ndarray) -> jnp.ndarray:
    """Computes the likelihoods P(g_values|watermarked).

    Args:
      g_values: g-values (values 0 or 1) of shape [batch_size, seq_len,
        watermarking_depth]

    Returns:
      p(g_values|watermarked) of shape [batch_size, seq_len,
      watermarking_depth].
    """
    p_one_unique_token, p_two_unique_tokens = self._compute_latents(g_values)

    # P(g_tl | watermarked) is equal to
    # 0.5 * [ (g_tl+0.5) * p_two_unique_tokens + p_one_unique_token].
    return 0.5 * ((g_values + 0.5) * p_two_unique_tokens + p_one_unique_token)


class LikelihoodModelUnwatermarked(nn.Module, LikelihoodModel):
  """Unwatermarked likelihood model for binary-valued g-values.

  This takes in g-values and returns p(g_values | not watermarked).
  """

  @nn.compact
  def __call__(self, g_values: jnp.ndarray) -> jnp.ndarray:
    """Computes the likelihoods P(g-values|not watermarked).

    Args:
      g_values: g-values (0 or 1 values) of shape [batch_size, seq_len,
        watermarking_depth, ...].

    Returns:
      Likelihoods of g-values given text is unwatermarked --
      p(g_values | not watermarked) of shape [batch_size, seq_len,
      watermarking_depth].
    """
    return 0.5 * jnp.ones_like(g_values)  # all g-values have prob 0.5.


def _compute_posterior(
    likelihoods_watermarked: jnp.ndarray,
    likelihoods_unwatermarked: jnp.ndarray,
    mask: jnp.ndarray,
    prior: float,
) -> jnp.ndarray:
  """Compute posterior P(w|g) given likelihoods, mask and prior.

  Args:
    likelihoods_watermarked: shape [batch, length, depth]. Likelihoods
      P(g_values|watermarked) of g-values under watermarked model.
    likelihoods_unwatermarked: shape [batch, length, depth]. Likelihoods
      P(g_values|unwatermarked) of g-values under unwatermarked model.
    mask: A binary array shape [batch, length] indicating which g-values should
      be used. g-values with mask value 0 are discarded.
    prior: Prior probability P(w) that the text is watermarked.

  Returns:
    Posterior probability P(watermarked|g_values), shape [batch].
  """
  mask = jnp.expand_dims(mask, -1)
  prior = jnp.clip(prior, a_min=1e-5, a_max=1 - 1e-5)
  log_likelihoods_watermarked = jnp.log(
      jnp.clip(likelihoods_watermarked, a_min=1e-30, a_max=float("inf"))
  )
  log_likelihoods_unwatermarked = jnp.log(
      jnp.clip(likelihoods_unwatermarked, a_min=1e-30, a_max=float("inf"))
  )
  log_odds = log_likelihoods_watermarked - log_likelihoods_unwatermarked

  # Sum relative surprisals (log odds) across all token positions and layers.
  relative_surprisal_likelihood = jnp.einsum(
      "i...->i", log_odds * mask
  )  # [batch_size].

  relative_surprisal_prior = jnp.log(prior) - jnp.log(1 - prior)

  # Combine prior and likelihood.
  relative_surprisal = (
      relative_surprisal_prior + relative_surprisal_likelihood
  )  # [batch_size]

  # Compute the posterior probability P(w|g) = sigmoid(relative_surprisal).
  return jax.nn.sigmoid(relative_surprisal)


class BayesianDetectorModule(nn.Module):
  """Bayesian classifier for watermark detection Flax Module.

  This detector uses Bayes' rule to compute a watermarking score, which is
  the posterior probability P(watermarked|g_values) that the text is
  watermarked, given its g_values.

  Note that this detector only works with Tournament-based watermarking using
  the Bernoulli(0.5) g-value distribution.
  """

  watermarking_depth: int  # The number of tournament layers.
  params: Mapping[str, Mapping[str, Any]] | None = None
  baserate: float = 0.5  # Prior probability P(w) that a text is watermarked.

  @property
  def score_type(self) -> ScoreType:
    return ScoreType.POSTERIOR

  def l2_loss(self) -> jnp.ndarray:
    return self.likelihood_model_watermarked.l2_loss()

  def setup(self):
    """Initializes the model parameters."""

    def _fetch_params():
      return {"params:": self.params["params"]["likelihood_model_watermarked"]}

    self.likelihood_model_watermarked = LikelihoodModelWatermarked(
        watermarking_depth=self.watermarking_depth,
        params=_fetch_params() if self.params is not None else None,
    )
    self.likelihood_model_unwatermarked = LikelihoodModelUnwatermarked()
    self.prior = self.param("prior", lambda *x: self.baserate, (1,))

  def __call__(
      self,
      g_values: jnp.ndarray,
      mask: jnp.ndarray,
  ) -> jnp.ndarray:
    """Computes the watermarked posterior P(watermarked|g_values).

    Args:
      g_values: g-values (with values 0 or 1) of shape [batch_size, seq_len,
        watermarking_depth, ...]
      mask: A binary array shape [batch_size, seq_len] indicating which g-values
        should be used. g-values with mask value 0 are discarded.

    Returns:
      P(watermarked | g_values), of shape [batch_size].
    """

    likelihoods_watermarked = self.likelihood_model_watermarked(g_values)
    likelihoods_unwatermarked = self.likelihood_model_unwatermarked(g_values)
    return _compute_posterior(
        likelihoods_watermarked, likelihoods_unwatermarked, mask, self.prior
    )

  def score(
      self,
      g_values: jnp.ndarray | Sequence[jnp.ndarray],
      mask: jnp.ndarray,
  ) -> jnp.ndarray:
    if self.params is None:
      raise ValueError("params must be set before calling score")
    return self.apply(self.params, g_values, mask, method=self.__call__)


def xentropy_loss(y: jnp.ndarray, y_pred: jnp.ndarray) -> jnp.ndarray:
  """Calculates cross entropy loss."""
  y_pred = jnp.clip(y_pred, 1e-5, 1 - 1e-5)
  return -jnp.mean((y * jnp.log(y_pred) + (1 - y) * jnp.log(1 - y_pred)))


def loss_fn(
    params: Mapping[str, Any],
    detector_inputs: Any,
    w_true: jnp.ndarray,
    l2_batch_weight: float,
    detector_module: BayesianDetectorModule,
) -> jnp.ndarray:
  """Calculates loss for a batch of data given parameters."""
  w_pred = detector_module.apply(
      params, *detector_inputs, method=detector_module.__call__
  )
  unweighted_l2 = detector_module.apply(params, method=detector_module.l2_loss)
  l2_loss = l2_batch_weight * unweighted_l2
  return xentropy_loss(w_true, w_pred) + l2_loss


def tpr_at_fpr(
    params: Mapping[str, Any],
    detector_inputs: Any,
    w_true: jnp.ndarray,
    minibatch_size,
    detector_module: BayesianDetectorModule,
    target_fpr: float = 0.01,
) -> jnp.ndarray:
  """Calculates TPR at FPR=target_fpr."""
  positive_idxs = w_true == 1
  negative_idxs = w_true == 0
  inds = jnp.arange(0, len(detector_inputs[0]), minibatch_size)
  w_preds = []
  for start in inds:
    end = start + minibatch_size
    detector_inputs_ = (
        detector_inputs[0][start:end],
        detector_inputs[1][start:end],
    )
    w_pred = detector_module.apply(
        params, *detector_inputs_, method=detector_module.__call__
    )
    w_preds.append(w_pred)
  w_pred = jnp.concatenate(w_preds, axis=0)
  positive_scores = w_pred[positive_idxs]
  negative_scores = w_pred[negative_idxs]
  fpr_threshold = jnp.percentile(negative_scores, 100 - target_fpr * 100)
  return jnp.mean(positive_scores >= fpr_threshold)


@enum.unique
class ValidationMetric(enum.Enum):
  """Direction along the z-axis."""

  TPR_AT_FPR = "tpr_at_fpr"
  CROSS_ENTROPY = "cross_entropy"


def train(
    *,
    detector_module: BayesianDetectorModule,
    g_values: jnp.ndarray,
    mask: jnp.ndarray,
    watermarked: jnp.ndarray,
    epochs: int = 250,
    learning_rate: float = 1e-3,
    minibatch_size: int = 64,
    seed: int = 0,
    l2_weight: float = 0.0,
    shuffle: bool = True,
    g_values_val: jnp.ndarray | None = None,
    mask_val: jnp.ndarray | None = None,
    watermarked_val: jnp.ndarray | None = None,
    verbose: bool = False,
    validation_metric: ValidationMetric = ValidationMetric.TPR_AT_FPR,
) -> tuple[Mapping[int, Mapping[str, PyTree]], float]:
  """Trains a Bayesian detector model.

  Args:
    detector_module: The detector module to train in-place.
    g_values: g-values of shape [num_train, seq_len, watermarking_depth].
    mask: A binary array shape [num_train, seq_len] indicating which g-values
      should be used. g-values with mask value 0 are discarded.
    watermarked: A binary array of shape [num_train] indicating whether the
      example is watermarked (0: unwatermarked, 1: watermarked).
    epochs: Number of epochs to train for.
    learning_rate: Learning rate for optimizer.
    minibatch_size: Minibatch size for training. Note that a minibatch requires
      ~ 32 * minibatch_size * seq_len * watermarked_depth * watermarked_depth
      bits of memory.
    seed: Seed for parameter initialization.
    l2_weight: Weight to apply to L2 regularization for delta parameters.
    shuffle: Whether to shuffle before training.
    g_values_val: Validation g-values of shape [num_val, seq_len,
      watermarking_depth].
    mask_val: Validation mask of shape [num_val, seq_len].
    watermarked_val: Validation watermark labels of shape [num_val].
    verbose: Boolean indicating verbosity of training. If true, the loss will be
      printed. Defaulted to False.
    validation_metric: validation metric to use.

  Returns:
    Tuple of
      training_history: Training history keyed by epoch number where the
      values are
        dictionaries containing the loss, validation loss, and model
        parameters,
        keyed by
        'loss', 'val_loss', and 'params', respectively.
      min_val_loss: Minimum validation loss achieved during training.
  """
  minibatch_inds = jnp.arange(0, len(g_values), minibatch_size)
  minibatch_inds_val = None
  if g_values_val is not None:
    minibatch_inds_val = jnp.arange(0, len(g_values_val), minibatch_size)

  rng = jax.random.PRNGKey(seed)
  param_rng, shuffle_rng = jax.random.split(rng)

  def coshuffle(*args):
    return [jax.random.permutation(shuffle_rng, x) for x in args]

  if shuffle:
    g_values, mask, watermarked = coshuffle(g_values, mask, watermarked)

  def update_fn_if_fpr_tpr(params):
    """Loss function for negative TPR@FPR=1% as the validation loss."""
    tpr_ = tpr_at_fpr(
        params=params,
        detector_inputs=(g_values_val, mask_val),
        w_true=watermarked_val,
        minibatch_size=minibatch_size,
        detector_module=detector_module,
    )
    return -tpr_

  n_minibatches = len(g_values) / minibatch_size
  l2_batch_weight_train = l2_weight / n_minibatches
  l2_batch_weight_val = 0.0
  loss_fn_train = functools.partial(
      loss_fn,
      l2_batch_weight=l2_batch_weight_train,
      detector_module=detector_module,
  )
  loss_fn_jitted_val = jax.jit(
      functools.partial(
          loss_fn,
          l2_batch_weight=l2_batch_weight_val,
          detector_module=detector_module,
      )
  )

  @jax.jit
  def update(gvalues, masks, labels, params, opt_state):
    loss_fn_partialed = functools.partial(
        loss_fn_train,
        detector_inputs=(gvalues, masks),
        w_true=labels,
    )
    loss, grads = jax.value_and_grad(loss_fn_partialed)(params)
    updates, opt_state = optimizer.update(grads, opt_state)
    params = optax.apply_updates(params, updates)
    return loss, params, opt_state

  def update_with_minibatches(gvalues, masks, labels, inds, params, opt_state):
    """Update params iff opt_state is not None and always returns the loss."""
    losses = []
    for start in inds:
      end = start + minibatch_size
      loss, params, opt_state = update(
          gvalues[start:end],
          masks[start:end],
          labels[start:end],
          params,
          opt_state,
      )
      losses.append(loss)
    loss = jnp.mean(jnp.array(losses))
    return loss, params, opt_state

  def validate_with_minibatches(gvalues, masks, labels, inds, params):
    """Update params iff opt_state is not None and always returns the loss."""
    losses = []
    for start in inds:
      end = start + minibatch_size
      loss = loss_fn_jitted_val(
          params,
          detector_inputs=(gvalues[start:end], masks[start:end]),
          w_true=labels[start:end],
      )
      losses.append(loss)
    return jnp.mean(jnp.array(losses))

  def update_fn(opt_state, params):
    """Updates the model parameters and returns the loss."""
    loss, params, opt_state = update_with_minibatches(
        g_values, mask, watermarked, minibatch_inds, params, opt_state
    )
    val_loss = None
    if g_values_val is not None:
      if validation_metric == ValidationMetric.TPR_AT_FPR:
        val_loss = update_fn_if_fpr_tpr(params)
      else:
        val_loss = validate_with_minibatches(
            g_values_val,
            mask_val,
            watermarked_val,
            minibatch_inds_val,
            params,
        )

    return opt_state, params, loss, val_loss

  params = detector_module.params
  if params is None:
    params = detector_module.init(param_rng, g_values[:1], mask[:1])

  optimizer = optax.adam(learning_rate=learning_rate)
  opt_state = optimizer.init(params)

  history = {}
  epochs_completed = 0
  while epochs_completed < epochs:
    opt_state, params, loss, val_loss = update_fn(opt_state, params)
    epochs_completed += 1

    history[epochs_completed] = {
        "loss": loss,
        "val_loss": val_loss,
        "params": params["params"],
    }
    if verbose:
      if val_loss is not None:
        print(
            f"Epoch {epochs_completed}: loss {loss} (train), {val_loss} (val)"
        )
      else:
        print(f"Epoch {epochs_completed}: loss {loss} (train)")
  detector_module.params = params
  val_loss = np.squeeze(
      np.array([history[epoch]["val_loss"] for epoch in range(1, epochs + 1)])
  )
  best_val_epoch = np.argmin(val_loss) + 1
  min_val_loss = val_loss[best_val_epoch - 1]
  print(f"Best val Epoch: {best_val_epoch}, min_val_loss: {min_val_loss}")
  detector_module.params = {"params": history[best_val_epoch]["params"]}
  return history, min_val_loss


class BayesianDetector:
  """Baysian detector class used for watermark detection."""

  detector_module: BayesianDetectorModule
  tokenizer: Any
  logits_processor: logits_processing.SynthIDLogitsProcessor

  def __init__(
      self,
      logits_processor: logits_processing.SynthIDLogitsProcessor,
      tokenizer: Any,
      params: Mapping[str, Mapping[str, Any]],
  ):
    self.detector_module = BayesianDetectorModule(
        watermarking_depth=len(logits_processor.keys),
        params=params,
    )
    self.logits_processor = logits_processor
    self.tokenizer = tokenizer

  def score(self, outputs: jnp.ndarray) -> jnp.ndarray:
    """Score the model output for possibility of being watermarked.

    Score is within [0, 1] where 0 is not watermarked and 1 is watermarked.

    Args:
      outputs: model output of shape [batch_size, output_len]

    Returns:
      scores of shape [batch_size]
    """
    # eos mask is computed, skip first ngram_len - 1 tokens
    # eos_mask will be of shape [batch_size, output_len]
    eos_token_mask = self.logits_processor.compute_eos_token_mask(
        input_ids=outputs,
        eos_token_id=self.tokenizer.eos_token_id,
    )[:, self.logits_processor.ngram_len - 1 :]

    # context repetition mask is computed
    context_repetition_mask = (
        self.logits_processor.compute_context_repetition_mask(
            input_ids=outputs,
        )
    )
    # context repitition mask shape [batch_size, output_len - (ngram_len - 1)]

    combined_mask = context_repetition_mask * eos_token_mask

    g_values = self.logits_processor.compute_g_values(
        input_ids=outputs,
    )
    # g values shape [batch_size, output_len - (ngram_len - 1), depth]
    return self.detector_module.score(
        g_values.cpu().numpy(), combined_mask.cpu().numpy()
    )

  @classmethod
  def process_raw_model_outputs(
      cls,
      *,
      tokenized_wm_outputs: Sequence[np.ndarray] | np.ndarray,
      tokenized_uwm_outputs: Sequence[np.ndarray] | np.ndarray,
      logits_processor: logits_processing.SynthIDLogitsProcessor,
      tokenizer: Any,
      torch_device: torch.device,
      test_size: float = 0.3,
      pos_truncation_length: int | None = 200,
      neg_truncation_length: int | None = 100,
      max_padded_length: int = 2300,
  ) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Process raw models outputs into inputs we can train.

    Args:
      tokenized_wm_outputs: tokenized outputs of watermarked data.
      tokenized_uwm_outputs: tokenized outputs of unwatermarked data.
      logits_processor: logits processor used for watermarking.
      tokenizer: tokenizer used for the model.
      torch_device: torch device to use.
      test_size: test size to use for train-test split.
      pos_truncation_length: Length to truncate wm outputs. If None, no
        truncation is applied.
      neg_truncation_length: Length to truncate uwm outputs. If None, no
        truncation is applied.
      max_padded_length: Length to pad truncated outputs so that all processed
        entries have same shape.

    Returns:
      Tuple of train_g_values, train_masks, train_labels, cv_g_values, cv_masks,
        cv_labels
    """
    # Split data into train and CV
    train_wm_outputs, cv_wm_outputs = model_selection.train_test_split(
        tokenized_wm_outputs, test_size=test_size
    )

    train_uwm_outputs, cv_uwm_outputs = model_selection.train_test_split(
        tokenized_uwm_outputs, test_size=test_size
    )

    # Process both train and CV data for training
    wm_masks_train, wm_g_values_train = process_outputs_for_training(
        [
            torch.tensor(outputs, device=torch_device, dtype=torch.long)
            for outputs in train_wm_outputs
        ],
        logits_processor=logits_processor,
        tokenizer=tokenizer,
        pos_truncation_length=pos_truncation_length,
        neg_truncation_length=neg_truncation_length,
        max_length=max_padded_length,
        is_pos=True,
        is_cv=False,
        torch_device=torch_device,
    )
    wm_masks_cv, wm_g_values_cv = process_outputs_for_training(
        [
            torch.tensor(outputs, device=torch_device, dtype=torch.long)
            for outputs in cv_wm_outputs
        ],
        logits_processor=logits_processor,
        tokenizer=tokenizer,
        pos_truncation_length=pos_truncation_length,
        neg_truncation_length=neg_truncation_length,
        max_length=max_padded_length,
        is_pos=True,
        is_cv=True,
        torch_device=torch_device,
    )
    uwm_masks_train, uwm_g_values_train = process_outputs_for_training(
        [
            torch.tensor(outputs, device=torch_device, dtype=torch.long)
            for outputs in train_uwm_outputs
        ],
        logits_processor=logits_processor,
        tokenizer=tokenizer,
        pos_truncation_length=pos_truncation_length,
        neg_truncation_length=neg_truncation_length,
        max_length=max_padded_length,
        is_pos=False,
        is_cv=False,
        torch_device=torch_device,
    )
    uwm_masks_cv, uwm_g_values_cv = process_outputs_for_training(
        [
            torch.tensor(outputs, device=torch_device, dtype=torch.long)
            for outputs in cv_uwm_outputs
        ],
        logits_processor=logits_processor,
        tokenizer=tokenizer,
        pos_truncation_length=pos_truncation_length,
        neg_truncation_length=neg_truncation_length,
        max_length=max_padded_length,
        is_pos=False,
        is_cv=True,
        torch_device=torch_device,
    )

    # We get list of data; here we concat all together to be passed to the
    # detector.
    wm_masks_train = torch.cat(wm_masks_train, dim=0)
    wm_g_values_train = torch.cat(wm_g_values_train, dim=0)
    wm_labels_train = torch.ones((wm_masks_train.shape[0],), dtype=torch.bool)
    wm_masks_cv = torch.cat(wm_masks_cv, dim=0)
    wm_g_values_cv = torch.cat(wm_g_values_cv, dim=0)
    wm_labels_cv = torch.ones((wm_masks_cv.shape[0],), dtype=torch.bool)

    uwm_masks_train = torch.cat(uwm_masks_train, dim=0)
    uwm_g_values_train = torch.cat(uwm_g_values_train, dim=0)
    uwm_labels_train = torch.zeros(
        (uwm_masks_train.shape[0],), dtype=torch.bool
    )
    uwm_masks_cv = torch.cat(uwm_masks_cv, dim=0)
    uwm_g_values_cv = torch.cat(uwm_g_values_cv, dim=0)
    uwm_labels_cv = torch.zeros((uwm_masks_cv.shape[0],), dtype=torch.bool)

    # Concat pos and negatives data together.
    train_g_values = (
        torch.cat((wm_g_values_train, uwm_g_values_train), dim=0).cpu().numpy()
    )
    train_labels = (
        torch.cat((wm_labels_train, uwm_labels_train), axis=0).cpu().numpy()
    )
    train_masks = (
        torch.cat((wm_masks_train, uwm_masks_train), axis=0).cpu().numpy()
    )

    cv_g_values = (
        torch.cat((wm_g_values_cv, uwm_g_values_cv), axis=0).cpu().numpy()
    )
    cv_labels = torch.cat((wm_labels_cv, uwm_labels_cv), axis=0).cpu().numpy()
    cv_masks = torch.cat((wm_masks_cv, uwm_masks_cv), axis=0).cpu().numpy()

    # Free up GPU memory.
    del (
        wm_g_values_train,
        wm_labels_train,
        wm_masks_train,
        wm_g_values_cv,
        wm_labels_cv,
        wm_masks_cv,
    )
    gc.collect()
    torch.cuda.empty_cache()

    # Shuffle data.
    train_g_values = jnp.squeeze(train_g_values)
    train_labels = jnp.squeeze(train_labels)
    train_masks = jnp.squeeze(train_masks)

    cv_g_values = jnp.squeeze(cv_g_values)
    cv_labels = jnp.squeeze(cv_labels)
    cv_masks = jnp.squeeze(cv_masks)

    shuffled_idx = list(range(train_g_values.shape[0]))
    shuffled_idx = np.array(shuffled_idx)
    np.random.shuffle(shuffled_idx)
    train_g_values = train_g_values[shuffled_idx]
    train_labels = train_labels[shuffled_idx]
    train_masks = train_masks[shuffled_idx]

    shuffled_idx = list(range(cv_g_values.shape[0]))
    shuffled_idx = np.array(shuffled_idx)
    np.random.shuffle(shuffled_idx)
    cv_g_values = cv_g_values[shuffled_idx]
    cv_labels = cv_labels[shuffled_idx]
    cv_masks = cv_masks[shuffled_idx]

    return (
        train_g_values,
        train_masks,
        train_labels,
        cv_g_values,
        cv_masks,
        cv_labels,
    )

  @classmethod
  def train_best_detector_given_g_values(
      cls,
      *,
      train_g_values: jnp.ndarray,
      train_masks: jnp.ndarray,
      train_labels: jnp.ndarray,
      cv_g_values: jnp.ndarray,
      cv_masks: jnp.ndarray,
      cv_labels: jnp.ndarray,
      logits_processor: logits_processing.SynthIDLogitsProcessor,
      tokenizer: Any,
      n_epochs: int = 50,
      learning_rate: float = 2.1e-2,
      l2_weights: np.ndarray = np.logspace(-3, -2, num=4),
      verbose: bool = False,
  ) -> tuple["BayesianDetector", float]:
    """Train best detector given g_values, mask and labels."""
    best_detector = None
    lowest_loss = float("inf")
    val_losses = []
    for l2_weight in l2_weights:
      detector_module = BayesianDetectorModule(
          watermarking_depth=len(logits_processor.keys),
      )
      _, min_val_loss = train(
          detector_module=detector_module,
          g_values=train_g_values,
          mask=train_masks,
          watermarked=train_labels,
          g_values_val=cv_g_values,
          mask_val=cv_masks,
          watermarked_val=cv_labels,
          learning_rate=learning_rate,
          l2_weight=l2_weight,
          epochs=n_epochs,
          verbose=verbose,
      )
      val_losses.append(min_val_loss)
      if min_val_loss < lowest_loss:
        lowest_loss = min_val_loss
        best_detector = detector_module

    return cls(logits_processor, tokenizer, best_detector.params), lowest_loss

  @classmethod
  def train_best_detector(
      cls,
      *,
      tokenized_wm_outputs: Sequence[np.ndarray] | np.ndarray,
      tokenized_uwm_outputs: Sequence[np.ndarray] | np.ndarray,
      logits_processor: logits_processing.SynthIDLogitsProcessor,
      tokenizer: Any,
      torch_device: torch.device,
      test_size: float = 0.3,
      pos_truncation_length: int | None = 200,
      neg_truncation_length: int | None = 100,
      max_padded_length: int = 2300,
      n_epochs: int = 50,
      learning_rate: float = 2.1e-2,
      l2_weights: np.ndarray = np.logspace(-3, -2, num=4),
      verbose: bool = False,
  ) -> tuple["BayesianDetector", float]:
    """Construct, train and return the best detector based on wm and uwm data.

    In practice, we have found that tuning pos_truncation_length,
    neg_truncation_length, n_epochs, learning_rate and l2_weights can help
    improve the performance of the detector. We reccommend tuning these
    parameters for your data.

    Args:
      tokenized_wm_outputs: tokenized outputs of watermarked data.
      tokenized_uwm_outputs: tokenized outputs of unwatermarked data.
      logits_processor: logits processor used for watermarking.
      tokenizer: tokenizer used for the model.
      torch_device: torch device to use.
      test_size: test size to use for train-test split.
      pos_truncation_length: Length to truncate wm outputs. If None, no
        truncation is applied.
      neg_truncation_length: Length to truncate uwm outputs. If None, no
        truncation is done.
      max_padded_length: Length to pad truncated outputs so that all processed
        entries have same shape.
      n_epochs: Number of epochs to train the detector.
      learning_rate: Learning rate to use for training the detector.
      l2_weights: L2 weights to use for training the detector.
      verbose: Whether to print training progress.

    Returns:
      Tuple of trained detector and loss achieved on CV data.
    """
    if torch_device.type in ("cuda", "tpu"):
      raise ValueError(
          "We have found the training unstable on CPUs; we are working on"
          " a fix. Use GPU or TPU for training."
      )
    (
        train_g_values,
        train_masks,
        train_labels,
        cv_g_values,
        cv_masks,
        cv_labels,
    ) = cls.process_raw_model_outputs(
        tokenized_wm_outputs=tokenized_wm_outputs,
        tokenized_uwm_outputs=tokenized_uwm_outputs,
        logits_processor=logits_processor,
        tokenizer=tokenizer,
        torch_device=torch_device,
        test_size=test_size,
        pos_truncation_length=pos_truncation_length,
        neg_truncation_length=neg_truncation_length,
        max_padded_length=max_padded_length,
    )

    return cls.train_best_detector_given_g_values(
        train_g_values=train_g_values,
        train_masks=train_masks,
        train_labels=train_labels,
        cv_g_values=cv_g_values,
        cv_masks=cv_masks,
        cv_labels=cv_labels,
        logits_processor=logits_processor,
        tokenizer=tokenizer,
        verbose=verbose,
        n_epochs=n_epochs,
        learning_rate=learning_rate,
        l2_weights=l2_weights,
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Hashing function implementation."""

import torch


def accumulate_hash(
    current_hash: torch.LongTensor,
    data: torch.LongTensor,
    multiplier: int = 6364136223846793005,
    increment: int = 1,
) -> torch.LongTensor:
  """Accumulate hash of data on current hash.

  Method uses adapted linear congruential generator with newlib/musl parameters.

  This function has following property -
  f(x, data[T]) = f(f(x, data[:T - 1]), data[T])

  This function expects current_hash.shape and data.shape[:-1] to
  match/broadcastable.

  Args:
    current_hash: (shape,)
    data: (shape, tensor_len)
    multiplier: (int) multiplier of linear congruential generator
    increment: (int) increment of linear congruential generator

  Returns:
    upadted hash (shape,)
  """
  for i in range(data.shape[-1]):
    current_hash = torch.add(current_hash, data[..., i])
    current_hash = torch.mul(current_hash, multiplier)
    current_hash = torch.add(current_hash, increment)
  return current_hash



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from collections.abc import Sequence

from absl.testing import absltest
from absl.testing import parameterized
import immutabledict
import numpy as np
import torch
import tqdm

from . import logits_processing
from . import g_value_expectations
from . import torch_testing


def test_mean_g_value_matches_theoretical(
    vocab_size: int,
    ngram_len: int,
    batch_size: int,
    keys: Sequence[int],
    atol: float,
    device: torch.device,
    num_leaves: int = 2,
) -> tuple[float, float, bool]:
  """Tests that the mean g-value is close to theoretical value.

  SynthIDLogitsProcessor is tested on its own using random input tokens.

  Args:
    vocab_size: vocab size of the model.
    ngram_len: length of the ngram.
    batch_size: batch size of the model.
    keys: keys used for watermarking.
    atol: absolute tolerance for the mean g-value.
    device: device to use for the test.
    num_leaves: number of children per node in the tournament tree.

  Returns:
    A tuple of mean g-value, the expected mean g-value and the boolean result
    of the test.
  """
  generator = torch.Generator(device=device).manual_seed(0)
  # Use 10**9 rather than vocab_size to ensure variety in (n-1)-grams.
  context = torch.randint(
      low=0,
      high=10**9,
      size=(batch_size, ngram_len - 1),
      dtype=torch.int64,
      generator=generator,
      device=device,
  )

  context_history_size = 1024
  logits_processor = logits_processing.SynthIDLogitsProcessor(
      ngram_len=ngram_len,
      keys=keys,
      sampling_table_size=2**16,
      sampling_table_seed=0,
      context_history_size=context_history_size,
      device=device,
      top_k=vocab_size,
      temperature=0.7,
      num_leaves=num_leaves,
  )

  scores = torch.ones(
      (batch_size, vocab_size),
      dtype=torch.float64,
      device=device,
  )
  # Init state of the logits processor.
  logits_processor.watermarked_call(context, scores)
  # insert context into the state.
  for idx in range(1, ngram_len - 1):
    _ = logits_processor.watermarked_call(context[:, :idx], scores)

  updated_scores, indices_mapping, _ = logits_processor.watermarked_call(
      context, scores
  )

  probs = torch.nn.functional.softmax(updated_scores, dim=1)
  generator = torch.Generator(device=device).manual_seed(0)
  next_tokens = torch.multinomial(
      probs,
      num_samples=1,
      generator=generator,
  )
  # Re-map to dense indices with indices_mapping.
  next_tokens = torch.vmap(torch.take, in_dims=0, out_dims=0)(
      indices_mapping, next_tokens
  )

  ngrams = torch.concat((context, next_tokens), dim=1)
  g_values = logits_processor.compute_g_values(ngrams)
  mean_g_values = g_values.mean(dtype=torch.float64, dim=(0, 1))

  expected_mean_g_value = g_value_expectations.expected_mean_g_value(
      vocab_size=vocab_size, num_leaves=num_leaves
  )
  is_close = torch.all(
      torch.isclose(
          mean_g_values,
          torch.tensor(
              expected_mean_g_value, dtype=torch.float64, device=device
          ),
          atol=atol,
          rtol=0,
      )
  )

  return mean_g_values, expected_mean_g_value, is_close


class LogitsProcessorCorrectnessTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name='vocab_size_10k_ngram_len_5_num_layers_3',
          ngram_len=5,
          vocab_size=10000,
          num_layers=3,
      ),
      dict(
          testcase_name='vocab_size_1k_ngram_len_10_num_layers_5',
          ngram_len=10,
          vocab_size=1000,
          num_layers=5,
      ),
  )
  def test_g_value_uniformity_for_random_ngrams(
      self, vocab_size, ngram_len, num_layers
  ):
    device = torch_testing.torch_device()
    watermarking_config = immutabledict.immutabledict({
        'ngram_len': ngram_len,
        'keys': np.random.randint(low=0, high=2**16, size=(num_layers,)),
        'sampling_table_size': 2**16,
        'sampling_table_seed': 0,
        'context_history_size': 512,
        'device': device,
    })
    batch_size = 100000
    torch.manual_seed(0)
    ngrams = torch.randint(
        low=0,
        high=vocab_size,
        size=(batch_size, ngram_len),
        device=device,
    )

    logits_processor = logits_processing.SynthIDLogitsProcessor(
        **watermarking_config, top_k=10, temperature=1.0
    )
    g_values = logits_processor.compute_g_values(ngrams)
    g_values_mean = torch.mean(torch.mean(g_values.float(), dim=0))
    self.assertAlmostEqual(g_values_mean, 0.5, delta=0.01)

  @parameterized.named_parameters(
      dict(
          testcase_name='vocab_size_10k',
          vocab_size=10000,
          num_layers=3,
      ),
      dict(
          testcase_name='vocab_size_1k',
          vocab_size=1000,
          num_layers=20,
      ),
  )
  def test_g_values_uniformity_across_vocab_size(self, vocab_size, num_layers):
    batch_size = 1000
    ngram_len = 5
    device = torch_testing.torch_device()
    watermarking_config = immutabledict.immutabledict({
        'ngram_len': ngram_len,
        'keys': np.random.randint(low=0, high=2**16, size=(num_layers,)),
        'sampling_table_size': 2**16,
        'sampling_table_seed': 0,
        'context_history_size': 512,
        'device': device,
    })
    n_minus_1_grams = torch.randint(
        low=0,
        high=vocab_size,
        size=(batch_size, watermarking_config['ngram_len'] - 1),
        device=device,
    )

    logits_processor = logits_processing.SynthIDLogitsProcessor(
        **watermarking_config, top_k=10, temperature=1.0
    )
    ngram_keys, _ = logits_processor._compute_keys(
        n_minus_1_grams,
        torch.stack(
            [torch.arange(vocab_size, device=device) for _ in range(batch_size)]
        ),
    )

    g_values = logits_processor.sample_g_values(ngram_keys)
    # g_values shape should be [batch_size, vocab_size, num_layers]
    g_values_mean = torch.mean(torch.mean(g_values.float(), dim=1))
    self.assertAlmostEqual(g_values_mean, 0.5, delta=0.001)

  def test_distributional_convergence(self):
    """Check if watermarked distribution converges to input distribution."""
    vocab_size = 2
    batch_size = 1500
    num_keys = 1000
    device = torch_testing.torch_device()
    temperature = 1.0

    updated_softmaxes = 0
    for _ in tqdm.tqdm(range(num_keys)):
      watermarking_config = immutabledict.immutabledict({
          'ngram_len': 5,
          'keys': np.random.randint(0, 10**9, size=(1,), dtype=np.int64),
          'sampling_table_size': 2**16,
          'sampling_table_seed': 0,
          'context_history_size': 1024,
          'device': device,
      })

      logits_processor = logits_processing.SynthIDLogitsProcessor(
          **watermarking_config,
          top_k=vocab_size,
          temperature=temperature,
          apply_top_k=False,
      )

      ngrams = torch.randint(
          low=0,
          high=vocab_size,
          size=(batch_size, watermarking_config['ngram_len']),
          device=device,
      )

      # Insert ngram-1 into logit_processor state.
      for idx in range(watermarking_config['ngram_len'] - 1):
        _ = logits_processor.watermarked_call(
            ngrams[:, :idx], torch.ones((batch_size, vocab_size), device=device)
        )

      scores = torch.ones((batch_size, vocab_size), device=device)
      updated_scores, _, _ = logits_processor.watermarked_call(ngrams, scores)
      updated_softmaxes += (
          torch.nn.functional.softmax(updated_scores, dim=1).cpu().numpy()
      )
    updated_softmaxes = np.mean(updated_softmaxes, axis=0) / num_keys
    for softmax in updated_softmaxes:
      self.assertAlmostEqual(softmax, 0.5, delta=0.002)

  @parameterized.named_parameters(
      dict(
          testcase_name='vocab_size_100_ngram_len_10_num_layers_1',
          vocab_size=2,
          ngram_len=10,
          num_layers=1,
          atol=0.01,
      ),
      dict(
          testcase_name='vocab_size_100_ngram_len_5_num_layers_1',
          vocab_size=100,
          ngram_len=5,
          num_layers=1,
          atol=0.01,
      ),
      dict(
          testcase_name='vocab_size_100_ngram_len_10_num_layers_2',
          vocab_size=100,
          ngram_len=10,
          num_layers=2,
          atol=0.02,
      ),
      dict(
          testcase_name='vocab_size_2_ngram_len_10_num_layers_1_num_leaves_3',
          vocab_size=2,
          ngram_len=10,
          num_layers=1,
          num_leaves=3,
          atol=0.02,
      ),
      dict(
          testcase_name='vocab_size_100_ngram_len_10_num_layers_1_num_leaves_3',
          vocab_size=100,
          ngram_len=10,
          num_layers=1,
          num_leaves=3,
          atol=0.02,
      ),
  )
  def test_bias_from_logits_processor(
      self, vocab_size, ngram_len, num_layers, atol, num_leaves: int = 2,
  ):
    """Check if watermarked distribution converges to input distribution."""
    device = torch_testing.torch_device()
    result = test_mean_g_value_matches_theoretical(
        vocab_size=vocab_size,
        ngram_len=ngram_len,
        batch_size=20_000,
        keys=[np.random.randint(0, 10**9) for _ in range(num_layers)],
        atol=atol,
        device=device,
        num_leaves=num_leaves,
    )
    self.assertTrue(result[2])


class LogitsProcessorTest(absltest.TestCase):

  def set_up_logits_processor(
      self,
      batch_size,
      sequence_len,
      num_layers,
      ngram_len,
      top_k,
      vocab_size,
  ):
    """Setup function for all the tests."""
    device = torch_testing.torch_device()
    watermarking_config = immutabledict.immutabledict({
        'ngram_len': ngram_len,
        'keys': np.random.randint(low=0, high=2**16, size=(num_layers,)),
        'sampling_table_size': 2**16,
        'sampling_table_seed': 0,
        'context_history_size': 512,
        'device': device,
    })
    logits_processor = logits_processing.SynthIDLogitsProcessor(
        **watermarking_config, top_k=top_k, temperature=1.0
    )
    sequences = torch.randint(
        low=0,
        high=vocab_size,
        size=(batch_size, sequence_len),
        device=device,
    )
    return logits_processor, sequences, device

  def test_compute_g_values_shape(self):
    batch_size, sequence_len, num_layers, ngram_len = 1000, 50, 3, 5
    logits_processor, sequences, _ = self.set_up_logits_processor(
        batch_size,
        sequence_len,
        num_layers,
        ngram_len,
        top_k=10,
        vocab_size=32,
    )
    g_values = logits_processor.compute_g_values(sequences)
    self.assertEqual(
        g_values.shape, (batch_size, sequence_len - (ngram_len - 1), num_layers)
    )

  def test_compute_context_repetition_mask_shape(self):
    batch_size, sequence_len, num_layers, ngram_len = 1000, 50, 3, 5
    logits_processor, sequences, _ = self.set_up_logits_processor(
        batch_size,
        sequence_len,
        num_layers,
        ngram_len,
        top_k=10,
        vocab_size=32,
    )
    context_mask = logits_processor.compute_context_repetition_mask(sequences)
    self.assertEqual(
        context_mask.shape, (batch_size, sequence_len - (ngram_len - 1))
    )

  def test_compute_eos_token_mask_shape(self):
    batch_size, sequence_len, num_layers, ngram_len = 1000, 50, 3, 5
    logits_processor, sequences, _ = self.set_up_logits_processor(
        batch_size,
        sequence_len,
        num_layers,
        ngram_len,
        top_k=10,
        vocab_size=32,
    )
    eos_mask = logits_processor.compute_eos_token_mask(
        sequences, eos_token_id=0
    )
    self.assertEqual(eos_mask.shape, (batch_size, sequence_len))

  def test_compute_ngram_keys_shape(self):
    batch_size, sequence_len, num_layers, ngram_len = 1000, 50, 3, 5
    vocab_size = 32
    logits_processor, _, device = self.set_up_logits_processor(
        batch_size,
        sequence_len,
        num_layers,
        ngram_len,
        top_k=10,
        vocab_size=vocab_size,
    )

    num_ngrams = 20
    ngrams = torch.randint(
        low=0,
        high=vocab_size,
        size=(batch_size, num_ngrams, ngram_len),
        device=device,
    )
    ngram_keys = logits_processor.compute_ngram_keys(ngrams)
    self.assertEqual(ngram_keys.shape, (batch_size, num_ngrams, num_layers))

  def test_watermarked_call_shape(self):
    batch_size, sequence_len, num_layers, ngram_len = 1000, 50, 3, 5
    vocab_size = 32
    top_k = 10
    logits_processor, sequences, device = self.set_up_logits_processor(
        batch_size,
        sequence_len,
        num_layers,
        ngram_len,
        top_k=top_k,
        vocab_size=vocab_size,
    )
    scores = torch.ones((batch_size, vocab_size), device=device)
    watermarked_scores, top_k_indices, original_scores = (
        logits_processor.watermarked_call(sequences, scores)
    )
    self.assertEqual(watermarked_scores.shape, (batch_size, top_k))
    self.assertEqual(top_k_indices.shape, (batch_size, top_k))
    self.assertEqual(original_scores.shape, (batch_size, top_k))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Expected g-value for watermarking."""


def expected_mean_g_value(
    vocab_size: int,
    num_leaves: int = 2,
) -> float:
  """Compute expected mean g-value after watermarking, assuming uniform LM dist.

  This is the theoretical expected value for a single-layer of tournament
  watermarking, using a Bernoulli(0.5) g-value distribution and N=num_leaves
  samples, assuming that the LM distribution p_LM is uniform.

  Args:
    vocab_size: The size of the vocabulary.
    num_leaves: Number of leaves per node in the tournament tree (N in the
      paper).

  Returns:
    The expected mean g-value for watermarked text.
  """
  if num_leaves == 2:
    # This equation is from Corollary 27 in Supplementary Information of paper,
    # in the case where p_LM is uniform.
    return 0.5 + 0.25 * (1 - (1 / vocab_size))
  elif num_leaves == 3:
    # This case can be derived from Theorem 25 in Supplementary Information of
    # the paper, in the case where N=3 and p_LM is uniform.
    return 7 / 8 - (3 / (8 * vocab_size))
  else:
    raise ValueError(
        f'Only 2 or 3 leaves are supported for the expected mean g-value'
        f' computation, but got {num_leaves}.'
    )





# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Code for Mean and Weighted Mean scoring functions."""

import jax.numpy as jnp


def mean_score(
    g_values: jnp.ndarray,
    mask: jnp.ndarray,
) -> jnp.ndarray:
  """Computes the Mean score.

  Args:
    g_values: g-values of shape [batch_size, seq_len, watermarking_depth].
    mask: A binary array shape [batch_size, seq_len] indicating which g-values
      should be used. g-values with mask value 0 are discarded.

  Returns:
    Mean scores, of shape [batch_size]. This is the mean of the unmasked
      g-values.
  """
  watermarking_depth = g_values.shape[-1]
  num_unmasked = jnp.sum(mask, axis=1)  # shape [batch_size]
  return jnp.sum(g_values * jnp.expand_dims(mask, 2), axis=(1, 2)) / (
      watermarking_depth * num_unmasked
  )


def weighted_mean_score(
    g_values: jnp.ndarray,
    mask: jnp.ndarray,
    weights: jnp.ndarray | None = None,
) -> jnp.ndarray:
  """Computes the Weighted Mean score.

  Args:
    g_values: g-values of shape [batch_size, seq_len, watermarking_depth].
    mask: A binary array shape [batch_size, seq_len] indicating which g-values
      should be used. g-values with mask value 0 are discarded.
    weights: array of non-negative floats, shape [watermarking_depth]. The
      weights to be applied to the g-values. If not supplied, defaults to
      linearly decreasing weights from 10 to 1.

  Returns:
    Weighted Mean scores, of shape [batch_size]. This is the mean of the
      unmasked g-values, re-weighted using weights.
  """
  watermarking_depth = g_values.shape[-1]

  if weights is None:
    weights = jnp.linspace(start=10, stop=1, num=watermarking_depth)

  # Normalise weights so they sum to watermarking_depth.
  weights *= watermarking_depth / jnp.sum(weights)

  # Apply weights to g-values.
  g_values *= jnp.expand_dims(weights, axis=(0, 1))

  num_unmasked = jnp.sum(mask, axis=1)  # shape [batch_size]
  return jnp.sum(g_values * jnp.expand_dims(mask, 2), axis=(1, 2)) / (
      watermarking_depth * num_unmasked
  )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import dataclasses

from absl.testing import absltest
import mock
import torch
import transformers
from transformers import utils as transformers_utils

from . import synthid_mixin
from . import logits_processing


@dataclasses.dataclass(frozen=True, kw_only=True)
class Config:

  is_encoder_decoder: bool = True


class TestSynthIDModel(synthid_mixin.SynthIDSparseTopKMixin):

  def __init__(self, config: Config):
    super().__init__()
    self.config = config

  def _get_initial_cache_position(self, input_ids, model_kwargs):
    model_kwargs["cache_position"] = torch.zeros(input_ids.shape[1])
    del input_ids
    return model_kwargs

  def prepare_inputs_for_generation(self, input_ids, **kwargs):
    return {
        "x": input_ids,
        **kwargs,
    }

  def __call__(
      self,
      x: torch.Tensor,
      **kwargs,
  ) -> transformers_utils.ModelOutput:
    return transformers_utils.ModelOutput(
        logits=torch.ones(x.shape[0], 5, 7),
    )


class SynthidMixinTest(absltest.TestCase):

  def test_sampling_from_mixin_includes_watermarking(self):
    old_watermarked_call = (
        logits_processing.SynthIDLogitsProcessor.watermarked_call
    )
    with mock.patch.object(
        logits_processing.SynthIDLogitsProcessor,
        "watermarked_call",
        autospec=True,
    ) as mock_watermarked_call:
      mock_watermarked_call.side_effect = old_watermarked_call
      synthid_model = TestSynthIDModel(config=Config())
      generation_config = transformers.GenerationConfig(
          top_k=5,
          temperature=0.5,
          do_sample=True,
      )

      logits_warper = synthid_model._get_logits_warper(generation_config)

      synthid_model._sample(
          input_ids=torch.ones(3, 11, dtype=torch.long),
          logits_processor=transformers.LogitsProcessorList(),
          stopping_criteria=transformers.StoppingCriteriaList(
              [lambda *_: True]
          ),
          generation_config=generation_config,
          synced_gpus=False,
          streamer=None,
          logits_warper=logits_warper,
      )

      self.assertEqual(mock_watermarked_call.call_count, 1)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""SynthID watermarked mixin class."""

from collections.abc import Mapping
from typing import Any

import immutabledict
import torch
import transformers

from . import logits_processing


DEFAULT_WATERMARKING_CONFIG = immutabledict.immutabledict({
    "ngram_len": 5,  # This corresponds to H=4 context window size in the paper.
    "keys": [
        654,
        400,
        836,
        123,
        340,
        443,
        597,
        160,
        57,
        29,
        590,
        639,
        13,
        715,
        468,
        990,
        966,
        226,
        324,
        585,
        118,
        504,
        421,
        521,
        129,
        669,
        732,
        225,
        90,
        960,
    ],
    "sampling_table_size": 2**16,
    "sampling_table_seed": 0,
    "context_history_size": 1024,
    "device": (
        torch.device("cuda:0")
        if torch.cuda.is_available()
        else torch.device("cpu")
    ),
})


class SynthIDSparseTopKMixin(transformers.GenerationMixin):
  """Mixin class of transfomers library with watermarking enabled."""

  def _construct_warper_list(
      self, extra_params: Mapping[str, Any]
  ) -> transformers.LogitsProcessorList:
    """Instantiate warpers list."""
    warpers = transformers.LogitsProcessorList()
    warpers.append(
        logits_processing.SynthIDLogitsProcessor(
            **DEFAULT_WATERMARKING_CONFIG, **extra_params
        )
    )
    return warpers

  def _get_logits_warper(
      self,
      generation_config: transformers.GenerationConfig,
      **unused_kw,
  ) -> transformers.LogitsProcessorList:
    """Constructs and returns a list of warpers.

    This overrides the base class's implementation to control how we apply top_k
    and temperature. Only the SynthIDLogitsProcessor warper is constructed that
    performs top_k and temperature scaling before applying watermark. This is
    to improve the latency impact by watermarking by only considering the top_k
    indices for watermarking.

    Args:
     generation_config: Config used for generation with this model.

    Returns:
     List of logits processors to be applied at inference time.
    """
    extra_params = {}

    # Add temperature to extra params
    if not (
        generation_config.temperature is not None
        and 0.0 <= generation_config.temperature <= 1.0
    ):
      raise ValueError(
          f"Invalid temperature {generation_config.temperature} when sampling"
          " with watermarking. Temperature should be between 0.0 and 1.0."
      )
    extra_params["temperature"] = generation_config.temperature

    # Add top_k to extra params.
    if not (
        generation_config.top_k is not None and generation_config.top_k >= 1
    ):
      raise ValueError(
          f"Invalid top_k {generation_config.top_k} when sampling with"
          " watermarking. Top_k should >= 1."
      )
    extra_params["top_k"] = generation_config.top_k

    return self._construct_warper_list(extra_params)

  def _sample(
      self,
      input_ids: torch.LongTensor,
      logits_processor: transformers.LogitsProcessorList,
      stopping_criteria: transformers.StoppingCriteriaList,
      generation_config: transformers.GenerationConfig,
      synced_gpus: bool,
      streamer: "transformers.BaseStreamer | None",
      logits_warper: transformers.LogitsProcessorList | None = None,
      **model_kwargs,
  ) -> transformers.generation.utils.GenerateNonBeamOutput | torch.LongTensor:
    r"""Sample sequence of tokens.

    Generates sequences of token ids for models with a language modeling head
    using **multinomial sampling** and
    can be used for text-decoder, text-to-text, speech-to-text, and
    vision-to-text models.

    This function is copied and changed minimally from the HuggingFace
    repository to support watermarking implementation.

    This overrides the base class implementation to achive watermarking of the
    logits before they are sampled. This is done specifically so as to preserve
    the top_k indices separately without making the logits dense with all the
    indices. This removes extra overhead of considering all possible indices for
    watermarking.

    Args:
        input_ids: The sequence used as a prompt for the generation.
        logits_processor: List of instances of class derived from
          [`LogitsProcessor`] used to modify the prediction scores of the
          language modeling head applied at each generation step.
        stopping_criteria: An instance of [`StoppingCriteriaList`]. List of
          instances of class derived from [`StoppingCriteria`] used to tell if
          the generation loop should stop.
        generation_config: The generation configuration to be used as
          parametrization of the decoding method.
        synced_gpus: Whether to continue running the while loop until max_length
          (needed for ZeRO stage 3)
        streamer: Streamer object that will be used to stream the generated
          sequences. Generated tokens are passed through
          `streamer.put(token_ids)` and the streamer is responsible for any
          further processing.
        logits_warper: List of instances of class derived from [`LogitsWarper`]
          used to warp the prediction score distribution of the language
          modeling head applied before multinomial sampling at each generation
          step. Only required with sampling strategies (i.e. `do_sample` is set
          in `generation_config`)
        **model_kwargs: Additional model specific kwargs will be forwarded to
          the `forward` function of the model. If model is an encoder-decoder
          model the kwargs should include `encoder_outputs`.

    Returns:
        A `torch.LongTensor` containing the generated tokens (default behaviour)
        or a
        [`~generation.GenerateDecoderOnlyOutput`] if
        `model.config.is_encoder_decoder=False` and
        `return_dict_in_generate=True` or a
        [`~generation.GenerateEncoderDecoderOutput`] if
        `model.config.is_encoder_decoder=True`.
    """
    # init values
    pad_token_id = generation_config.pad_token_id
    output_attentions = generation_config.output_attentions
    output_hidden_states = generation_config.output_hidden_states
    output_scores = generation_config.output_scores
    output_logits = generation_config.output_logits
    return_dict_in_generate = generation_config.return_dict_in_generate
    has_eos_stopping_criteria = any(
        hasattr(criteria, "eos_token_id") for criteria in stopping_criteria
    )
    do_sample = generation_config.do_sample
    if do_sample and not isinstance(
        logits_warper, transformers.LogitsProcessorList
    ):
      raise ValueError(
          "`do_sample` is set to `True`, `logits_warper` must be a"
          f" `LogitsProcessorList` instance (it is {logits_warper})."
      )

    # init attention / hidden states / scores tuples
    scores = () if (return_dict_in_generate and output_scores) else None
    raw_logits = () if (return_dict_in_generate and output_logits) else None
    decoder_attentions = (
        () if (return_dict_in_generate and output_attentions) else None
    )
    cross_attentions = (
        () if (return_dict_in_generate and output_attentions) else None
    )
    decoder_hidden_states = (
        () if (return_dict_in_generate and output_hidden_states) else None
    )

    # if model is an encoder-decoder, retrieve encoder attention weights and
    # hidden states
    encoder_attentions = None
    encoder_hidden_states = None
    if return_dict_in_generate and self.config.is_encoder_decoder:  # pytype: disable=attribute-error
      encoder_attentions = (
          model_kwargs["encoder_outputs"].get("attentions")
          if output_attentions
          else None
      )
      encoder_hidden_states = (
          model_kwargs["encoder_outputs"].get("hidden_states")
          if output_hidden_states
          else None
      )

    # keep track of which sequences are already finished
    batch_size = input_ids.shape[0]
    this_peer_finished = False
    unfinished_sequences = torch.ones(
        batch_size, dtype=torch.long, device=input_ids.device
    )
    model_kwargs = self._get_initial_cache_position(input_ids, model_kwargs)  # pytype: disable=attribute-error

    while self._has_unfinished_sequences(  # pytype: disable=attribute-error
        this_peer_finished, synced_gpus, device=input_ids.device
    ):
      # prepare model inputs
      model_inputs = self.prepare_inputs_for_generation(  # pytype: disable=attribute-error
          input_ids, **model_kwargs
      )

      # forward pass to get next token
      outputs = self(  # pytype: disable=not-callable
          **model_inputs,
          return_dict=True,
          output_attentions=output_attentions,
          output_hidden_states=output_hidden_states,
      )

      if synced_gpus and this_peer_finished:
        continue  # don't waste resources running the code we don't need

      # Clone is needed to avoid keeping a hanging ref to outputs.logits which
      # may be very large for first iteration (the clone itself is always small)
      next_token_logits = outputs.logits[:, -1, :].clone()

      # pre-process distribution
      next_token_scores = logits_processor(input_ids, next_token_logits)
      indices_mapping = None
      unwatermarked_scores = None
      if do_sample:
        *regular_warpers, watermarking_logits_warper = logits_warper
        if not isinstance(
            watermarking_logits_warper,
            logits_processing.SynthIDLogitsProcessor,
        ):
          raise ValueError(
              "SynthIDLogitsProcessor should be the final warper in the list"
              " while watermarking."
          )
        for logit_warper in regular_warpers:
          next_token_scores = logit_warper(input_ids, next_token_scores)
        # Watermark final scores with sparse top_k.
        next_token_scores, indices_mapping, unwatermarked_scores = (
            watermarking_logits_warper.watermarked_call(
                input_ids, next_token_scores
            )
        )

      # token selection
      if do_sample:
        probs = torch.nn.functional.softmax(next_token_scores, dim=-1)
        next_tokens = torch.multinomial(probs, num_samples=1).squeeze(1)
      else:
        next_tokens = torch.argmax(next_token_scores, dim=-1)

      # Store scores, attentions and hidden_states when required
      if return_dict_in_generate:
        if output_scores:
          assert unwatermarked_scores is not None
          score = torch.gather(
              -torch.log(torch.nn.Softmax(dim=1)(unwatermarked_scores)),
              1,
              next_tokens[:, None],
          )
          scores += (score,)
        if output_logits:
          raw_logits += (next_token_logits,)
        if output_attentions:
          decoder_attentions += (
              (outputs.decoder_attentions,)
              if self.config.is_encoder_decoder  # pytype: disable=attribute-error
              else (outputs.attentions,)
          )
          if self.config.is_encoder_decoder:  # pytype: disable=attribute-error
            cross_attentions += (outputs.cross_attentions,)

        if output_hidden_states:
          decoder_hidden_states += (
              (outputs.decoder_hidden_states,)
              if self.config.is_encoder_decoder  # pytype: disable=attribute-error
              else (outputs.hidden_states,)
          )

      assert indices_mapping is not None
      # re-mapping to dense indices with indices_mapping
      next_tokens = torch.vmap(torch.take, in_dims=0, out_dims=0)(
          indices_mapping, next_tokens
      )

      # finished sentences should have their next token be a padding token
      if has_eos_stopping_criteria:
        next_tokens = next_tokens * unfinished_sequences + pad_token_id * (
            1 - unfinished_sequences
        )

      # update generated ids, model inputs, and length for next step
      input_ids = torch.cat([input_ids, next_tokens[:, None]], dim=-1)
      if streamer is not None:
        streamer.put(next_tokens.cpu())
      model_kwargs = self._update_model_kwargs_for_generation(  # pytype: disable=attribute-error
          outputs,
          model_kwargs,
          is_encoder_decoder=self.config.is_encoder_decoder,  # pytype: disable=attribute-error
      )

      unfinished_sequences = unfinished_sequences & ~stopping_criteria(
          input_ids, scores
      )
      this_peer_finished = unfinished_sequences.max() == 0

      # This is needed to properly delete outputs.logits which may be very large
      # for first iteration. Otherwise a reference to outputs is kept which
      # keeps the logits alive in the next iteration
      del outputs

    if streamer is not None:
      streamer.end()

    if return_dict_in_generate:
      if self.config.is_encoder_decoder:  # pytype: disable=attribute-error
        return transformers.generation.utils.GenerateEncoderDecoderOutput(
            sequences=input_ids,
            scores=scores,
            logits=raw_logits,
            encoder_attentions=encoder_attentions,
            encoder_hidden_states=encoder_hidden_states,
            decoder_attentions=decoder_attentions,
            cross_attentions=cross_attentions,
            decoder_hidden_states=decoder_hidden_states,
            past_key_values=model_kwargs.get("past_key_values"),
        )
      else:
        return transformers.generation.utils.GenerateDecoderOnlyOutput(
            sequences=input_ids,
            scores=scores,
            logits=raw_logits,
            attentions=decoder_attentions,
            hidden_states=decoder_hidden_states,
            past_key_values=model_kwargs.get("past_key_values"),
        )
    else:
      return input_ids


class SynthIDGPT2LMHeadModel(
    SynthIDSparseTopKMixin, transformers.GPT2LMHeadModel
):
  pass


class SynthIDGemmaForCausalLM(
    SynthIDSparseTopKMixin, transformers.GemmaForCausalLM
):
  pass
